'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var helpers = require('./helpers.js');
var constants = require('./constants.js');
var formControl = require('./formControl.js');
var list = require('./list.js');
var token = require('./token.js');
var attribute = require('./attribute.js');
var context = require('./context.js');
var template = require('./template.js');
var partial = require('./partial.js');

const render = (
  target,
  { getState, dispatch, refs },
  template$1,
  updatedCallback,
  beforeMountCallback
) => {
  let observer = () => {
    let subscribers = new Set();
    return {
      publish: (cb) => {
        for (let fn of subscribers) {
          fn();
        }
        cb?.();
      },
      subscribe(fn) {
        subscribers.add(fn);
      },
    }
  };

  const createSubscription = {
    [constants.TEXT]: ({ value, node, context }, { getState }) => {
      return {
        handler: () => {
          let state = context ? context.wrap(getState()) : getState();
          let a = node.textContent;
          let b = token.getValueFromParts(state, token.getParts(value));

          if (a !== b) node.textContent = b;
        },
      }
    },
    [constants.ATTRIBUTE]: ({ value, node, name, context }, { getState }) => {
      return {
        handler: () => {
          let state = context ? context.wrap(getState()) : getState();
          let b = token.getValueFromParts(state, token.getParts(value));

          attribute.applyAttribute(node, name, b);

          if (node.nodeName === "OPTION") {
            let path = node.parentNode.getAttribute("name");
            let selected = helpers.getValueAtPath(path, state);
            node.selected = selected === b;
          }
        },
      }
    },
    [constants.INPUT]: ({ node, path, context }, { getState, dispatch }) => {
      let eventType = ["radio", "checkbox"].includes(node.type)
        ? "click"
        : "input";

      node.addEventListener(eventType, () => {
        let value =
          node.getAttribute("type") === "checkbox" ? node.checked : node.value;

        if (value.trim?.().length && !isNaN(value)) value = +value;

        if (context) {
          let state = context.wrap(getState());
          helpers.setValueAtPath(path, value, state);
          dispatch({
            type: "MERGE",
            payload: state,
          });
        } else {
          dispatch({
            type: "SET",
            payload: {
              name: path,
              value,
              context,
            },
          });
        }
      });

      return {
        handler: () => {
          let state = context ? context.wrap(getState()) : getState();
          formControl.updateFormControl(node, helpers.getValueAtPath(path, state));
        },
      }
    },
    [constants.EVENT]: (
      { node, eventType, actionType, context },
      { dispatch, getState }
    ) => {
      node.addEventListener(eventType, (event) => {
        let action = {
          type: actionType,
          event,
          scope: context ? context.wrap(getState()) : getState(),
        };

        dispatch(action);
      });
      return {
        handler: () => {},
      }
    },
    [constants.CONDITIONAL]: ({ node, expression, context, map }, { getState }) => {
      return {
        handler: () => {
          let state = context ? context.wrap(getState()) : getState();
          let shouldMount = token.getValueFromParts(
            state,
            token.getParts(helpers.wrapToken(expression))
          );
          let isMounted = node.dataset.mounted === "1";

          if (shouldMount && !isMounted) {
            // MOUNT
            let frag = helpers.fragmentFromTemplate(node);
            helpers.walk(frag.firstChild, bindAll(map, null, context));
            node.after(frag);
            node.dataset.mounted = "1";
          } else if (!shouldMount && isMounted) {
            // UNMOUNT
            node.nextSibling.remove();
            delete node.dataset.mounted;
          }
        },
        // pickupNode,
      }
    },
    [constants.REPEAT]: (
      { node, context: context$1, map, path, identifier, index, key, blockIndex, hydrate },
      { getState }
    ) => {
      let oldValue;
      node.$t = blockIndex - 1;

      let pickupNode = node.nextSibling;

      const initialiseBlock = (rootNode, i, k, exitNode) => {
        helpers.walk(
          rootNode,
          multi(
            (node) => {
              if (node === exitNode) return false
            },
            bindAll(
              map,
              hydrate,
              context.createContext(
                (context$1?.get() || []).concat({
                  path,
                  identifier,
                  key,
                  index,
                  i,
                  k,
                })
              )
            ),
            (child) => (child.$t = blockIndex)
          )
        );
      };

      function firstChild(v) {
        return (v.nodeType === v.DOCUMENT_FRAGMENT_NODE && v.firstChild) || v
      }

      const createListItem = (datum, i) => {
        let k = datum[key];
        let frag = helpers.fragmentFromTemplate(node);

        initialiseBlock(firstChild(frag), i, k);
        return frag
      };

      if (hydrate) {
        let x = helpers.getValueAtPath(path, getState());
        let blocks = list.getBlocks(node);

        blocks.forEach((block, i) => {
          let datum = x[i];
          let k = datum?.[key];
          initialiseBlock(block[0], i, k, helpers.last(block).nextSibling);
        });

        pickupNode = helpers.last(helpers.last(blocks))?.nextSibling;
      }

      return {
        handler: () => {
          let state = context$1 ? context$1.wrap(getState()) : getState();

          const newValue = Object.entries(helpers.getValueAtPath(path, state) || []);
          const delta = list.compareKeyedLists(key, oldValue, newValue);

          if (delta) {
            list.updateList(node, delta, newValue, createListItem);
          }
          oldValue = newValue.slice(0);
        },
        pickupNode,
      }
    },
  };

  const mediator = () => {
    const o = observer();
    return {
      bind(v) {
        let s = createSubscription[v.type](v, { getState, dispatch });

        o.subscribe(s.handler);
        return s
      },

      update(cb) {
        return o.publish(cb)
      },
    }
  };

  const { bind, update } = mediator();

  let blockCount = 0;

  function compareAttributes(a, b) {
    if (b.name === ":if") {
      return -1
    }
    if (a.name === ":if") {
      return 1
    }
    return 0
  }

  const parse = (frag) => {
    let index = 0;
    let map = {};

    helpers.walk(frag, (node) => {
      let x = [];
      let pickupNode;

      switch (node.nodeType) {
        case node.TEXT_NODE: {
          let value = node.textContent;
          if (token.hasMustache(value)) {
            x.push({
              type: constants.TEXT,
              value: value.trim(),
            });
          }
          break
        }
        case node.ELEMENT_NODE: {
          if (node.nodeName in partial.partials) {
            node.innerHTML = partial.partials[node.nodeName];
            break
          }

          let each = list.parseEach(node);

          if (each) {
            let ns = node.namespaceURI;
            let m;

            node.removeAttribute(":each");
            node = template.convertToTemplate(node);

            if (ns.endsWith("/svg")) {
              m = parse(node.firstChild);
            } else {
              m = parse(node.content.firstChild);
            }

            pickupNode = node.nextSibling;

            x.push({
              type: constants.REPEAT,
              map: m,
              blockIndex: blockCount++,
              ...each,
            });

            break
          }

          let attrs = [...node.attributes].sort(compareAttributes);
          let i = attrs.length;
          while (i--) {
            let { name, value } = attrs[i];

            if (
              name === ":name" &&
              value &&
              (node.nodeName === "INPUT" ||
                node.nodeName === "SELECT" ||
                node.nodeName === "TEXTAREA")
            ) {
              x.push({
                type: constants.INPUT,
                path: value,
              });

              node.removeAttribute(name);
              node.setAttribute("name", value);
            } else if (name === ":if") {
              node.removeAttribute(name);
              node = template.convertToTemplate(node);
              pickupNode = node.nextSibling;

              x.push({
                type: constants.CONDITIONAL,
                expression: value,
                map: parse(node.content?.firstChild || node.firstChild),
              });
            } else if (name === ":ref") {
              node.removeAttribute(name);
              node.setAttribute("ref", value);
              Object.defineProperty(refs, value, {
                get() {
                  return document.querySelector(`[ref="${value}"]`)
                },
              });
            } else if (name.startsWith(":on")) {
              node.removeAttribute(name);
              let eventType = name.split(":on")[1];
              x.push({
                type: constants.EVENT,
                eventType,
                actionType: value,
              });
            } else if (name.startsWith(":")) {
              let prop = name.slice(1);

              let v = value || prop;

              if (!v.includes("{{")) v = `{{${v}}}`;

              x.push({
                type: constants.ATTRIBUTE,
                name: prop,
                value: v,
              });
              node.removeAttribute(name);
            }
          }
        }
      }
      if (x.length) {
        map[index] = x;
      }
      index++;
      return pickupNode
    });

    return map
  };

  const multi =
    (...fns) =>
    (...args) => {
      for (let fn of fns) {
        let v = fn(...args);
        if (v === false) return false
      }
    };

  const bindAll = (bMap, hydrate = 0, context) => {
    let index = 0;
    return (node) => {
      let k = index;
      let p;

      if (k in bMap) {
        bMap[k].forEach((v) => {
          let x = bind({
            ...v,
            node,
            context,
            hydrate,
          });
          p = x.pickupNode;
        });
        node.$i = index;
      }

      index++;
      return p
    }
  };

  let frag = helpers.fragmentFromTemplate(template$1);
  let map = parse(frag);
  let ds = helpers.getDataScript(target);

  if (ds) {
    helpers.walk(target, bindAll(map, 1));
  } else {
    helpers.walk(frag, bindAll(map));
    beforeMountCallback?.(frag);
    update();
    target.prepend(frag);
  }

  return helpers.debounce(() => update(updatedCallback))
};

exports.render = render;
