const wrapToken = (v) => {
  v = v.trim();
  if (v.startsWith("{{")) return v
  return `{{${v}}}`
};

const last = (v = []) => v[v.length - 1];

const isWhitespace = (node) => {
  return node.nodeType === node.TEXT_NODE && node.nodeValue.trim() === ""
};

const walk = (node, callback, deep = true) => {
  if (!node) return
  // if (node.matches?.(`script[type="application/synergy"]`))
  //   return walk(node.nextSibling, callback, deep)
  if (!isWhitespace(node)) {
    let v = callback(node);
    if (v === false) return
    if (v?.nodeName) return walk(v, callback, deep)
  }
  if (deep) walk(node.firstChild, callback, deep);
  walk(node.nextSibling, callback, deep);
};

const transformBrackets = (str = "") => {
  let parts = str.split(/(\[[^\]]+\])/).filter((v) => v);
  return parts.reduce((a, part) => {
    let v = part.charAt(0) === "[" ? "." + part.replace(/\./g, ":") : part;
    return a + v
  }, "")
};

const getTarget = (path, target) => {
  let parts = transformBrackets(path)
    .split(".")
    .map((k) => {
      if (k.charAt(0) === "[") {
        let p = k.slice(1, -1).replace(/:/g, ".");
        return getValueAtPath(p, target)
      } else {
        return k
      }
    });

  let t =
    parts.slice(0, -1).reduce((o, k) => {
      return o && o[k]
    }, target) || target;
  return [t, last(parts)]
};

const getValueAtPath = (path, target) => {
  let [a, b] = getTarget(path, target);
  let v = a?.[b];
  if (typeof v === "function") return v.bind(a)
  return v
};

const setValueAtPath = (path, value, target) => {
  let [a, b] = getTarget(path, target);
  return (a[b] = value)
};

const fragmentFromTemplate = (v) => {
  if (typeof v === "string") {
    if (v.charAt(0) === "#") {
      v = document.querySelector(v);
    } else {
      let tpl = document.createElement("template");
      tpl.innerHTML = v.trim();
      return tpl.content.cloneNode(true)
    }
  }
  if (v.nodeName === "TEMPLATE") return v.cloneNode(true).content
  if (v.nodeName === "defs") return v.firstElementChild.cloneNode(true)
};

const debounce = (fn) => {
  let wait = false;
  let invoke = false;
  return () => {
    if (wait) {
      invoke = true;
    } else {
      wait = true;
      fn();
      requestAnimationFrame(() => {
        if (invoke) fn();
        wait = false;
      });
    }
  }
};

const isPrimitive = (v) => v === null || typeof v !== "object";

const typeOf = (v) =>
  Object.prototype.toString.call(v).match(/\s(.+[^\]])/)[1];

const pascalToKebab = (string) =>
  string.replace(/[\w]([A-Z])/g, function (m) {
    return m[0] + "-" + m[1].toLowerCase()
  });

const kebabToPascal = (string) =>
  string.replace(/[\w]-([\w])/g, function (m) {
    return m[0] + m[2].toUpperCase()
  });

const applyAttribute = (node, name, value) => {
  name = pascalToKebab(name);

  if (typeof value === "boolean") {
    if (name.startsWith("aria-")) {
      value = "" + value;
    } else if (value) {
      value = "";
    }
  }

  if (typeof value === "string" || typeof value === "number") {
    node.setAttribute(name, value);
  } else {
    node.removeAttribute(name);
  }
};

const attributeToProp = (k, v) => {
  let name = kebabToPascal(k);
  if (v === "") v = true;
  if (k.startsWith("aria-")) {
    if (v === "true") v = true;
    if (v === "false") v = false;
  }
  return {
    name,
    value: v,
  }
};

function getDataScript(node) {
  return node.querySelector(
    `script[type="application/synergy"][id="${node.nodeName}"]`
  )
}

function createDataScript(node) {
  let ds = document.createElement("script");
  ds.setAttribute("type", "application/synergy");
  ds.setAttribute("id", node.nodeName);
  node.append(ds);
  return ds
}

function systemReducer(state, action) {
  switch (action.type) {
    case "SET": {
      const { name, value } = action.payload;

      let o = { ...state };
      setValueAtPath(name, value, o);

      return o
    }
    case "MERGE": {
      return {
        ...state,
        ...action.payload,
      }
    }
  }
}

function configure({
  update = {},
  middleware = [],
  state: initialState = {},
  getState: getStateWrapper = (v) => v,
}) {
  let subscribers = [];
  let state;
  let onChangeCallback = () => {};

  function updateState(o) {
    state = getStateWrapper({ ...o });
    onChangeCallback();
  }

  updateState(initialState);

  const refs = {};

  function getState() {
    return { ...state }
  }

  function subscribe(fn) {
    subscribers.push(fn);
  }

  function updated() {
    subscribers.forEach((fn) => fn());
    subscribers = [];
  }

  function onChange(fn) {
    onChangeCallback = fn;
  }

  function dispatch(action) {
    const { type } = action;

    if (type === "SET" || type === "MERGE") {
      updateState(systemReducer(getState(), action));
    } else {
      const done = (action) => {
        if (action.type in update) {
          updateState(update[action.type](getState(), action));
        }
        return {
          then: (fn) =>
            new Promise((resolve) => {
              subscribe(() => {
                fn();
                resolve();
              });
            }),
        }
      };

      middleware[action.type]?.(action, done, {
        getState,
        dispatch,
        refs,
      }) || done(action);
    }
  }

  return {
    dispatch, // dispatch an action to the reducers
    getState, // optionally provide a wrapper function to derive additional properties in state
    onChange, // use this callback to update your UI whenever state changes
    updated, // call this once you've updated the UI so that all subscribers will be invoked and then removed
    refs, // an empty object that you can attach element refs to (supplied on object passed as the third argument to middleware functions)
  }
}

const TEXT = 1;
const ATTRIBUTE = 2;
const INPUT = 3;
const EVENT = 4;
const REPEAT = 5;
const CONDITIONAL = 6;

const updateFormControl = (node, value) => {
  if (node.nodeName === "SELECT") {
    Array.from(node.querySelectorAll("option")).forEach((option) => {
      option.selected = value.includes(option.value);
    });
    return
  }

  let checked;

  switch (node.getAttribute("type")) {
    case "checkbox":
      checked = value;
      if (node.checked === checked) break
      if (checked) {
        node.setAttribute("checked", "");
      } else {
        node.removeAttribute("checked");
      }
      break
    case "radio":
      checked = value === node.getAttribute("value");
      if (node.checked === checked) break
      node.checked = checked;
      if (checked) {
        node.setAttribute("checked", "");
      } else {
        node.removeAttribute("checked");
      }
      break
    default:
      if (node.value !== value) {
        if (!isNaN(value) && node.value.slice(-1) === ".") break
        node.setAttribute("value", (node.value = value ?? ""));
      }

      break
  }
};

function parseEach(node) {
  let each = node.getAttribute(":each");
  let m = each?.match(/(.+)\s+in\s+(.+)/);
  if (!m) {
    if (!each) return m
    return {
      path: each.trim(),
      key: node.getAttribute(":key"),
    }
  }
  let [_, left, right] = m;
  let parts = left.match(/\(([^\)]+)\)/);
  let [a, b] = (parts ? parts[1].split(",") : [left]).map((v) => v.trim());

  return {
    path: right.trim(),
    identifier: b ? b : a,
    index: b ? a : b,
    key: node.getAttribute(":key"),
  }
}

const getBlockSize = (template) => {
  let i = 0;
  walk(template.content?.firstChild || template.firstChild, () => i++, false);
  return i
};

const nextNonWhitespaceSibling = (node) => {
  return isWhitespace(node.nextSibling)
    ? nextNonWhitespaceSibling(node.nextSibling)
    : node.nextSibling
};

const getBlockFragments = (template, numBlocks) => {
  let blockSize = getBlockSize(template);

  let r = [];
  if (numBlocks) {
    while (numBlocks--) {
      let f = document.createDocumentFragment();
      let n = blockSize;
      while (n--) {
        f.appendChild(nextNonWhitespaceSibling(template));
      }
      r.push(f);
    }
  }
  return r
};

const getBlocks = (template) => {
  let numBlocks = +(template.dataset.length || 0);
  let blockSize = getBlockSize(template);
  let r = [];
  let node = template;
  if (numBlocks) {
    while (numBlocks--) {
      let f = [];
      let n = blockSize;
      while (n--) {
        node = nextNonWhitespaceSibling(node);
        f.push(node);
      }
      r.push(f);
    }
  }
  return r
};

const compareKeyedLists = (key, a = [], b = []) => {
  let delta = b.map(([k, item]) =>
    !key ? (k in a ? k : -1) : a.findIndex(([_, v]) => v[key] === item[key])
  );
  if (a.length !== b.length || !delta.every((a, b) => a === b)) return delta
};

function lastChild(v) {
  return (v.nodeType === v.DOCUMENT_FRAGMENT_NODE && v.lastChild) || v
}

const updateList = (template, delta, entries, createListItem) => {
  let n = +(template.dataset.length || 0);

  const unchanged = delta.length === n && delta.every((a, b) => a == b);

  if (unchanged) return

  let blocks = getBlockFragments(template, n);
  let t = template;

  delta.forEach((i, newIndex) => {
    let frag =
      i === -1
        ? createListItem(entries[newIndex][1], entries[newIndex][0])
        : blocks[i];
    let x = lastChild(frag);
    t.after(frag);
    t = x;
  });

  template.dataset.length = delta.length;
};

const VALUE = 1;
const KEY = 2;
const FUNCTION = 3;

const hasMustache = (v) => v.match(/({{[^{}]+}})/);

const getParts = (value) =>
  value
    .trim()
    .split(/({{[^{}]+}})/)
    .filter((v) => v)
    .map((value) => {
      let match = value.match(/{{([^{}]+)}}/);

      if (!match)
        return {
          type: VALUE,
          value,
          negated: false,
        }

      value = match[1].trim();
      let negated = value.charAt(0) === "!";
      if (negated) value = value.slice(1);

      return {
        type: KEY,
        value,
        negated,
      }
    });

const getValueFromParts = (target, parts) => {
  return parts.reduce((a, part) => {
    let { type, value, negated } = part;

    let v;

    if (type === VALUE) v = value;
    if (type === KEY) {
      v = getValueAtPath(value, target);
    }
    if (type === FUNCTION) {
      let args = part.args.map((value) => getValueAtPath(value, target));

      v = getValueAtPath(part.method, target)?.(...args);
    }

    if (negated) v = !v;

    return a || a === 0 ? a + v : v
  }, "")
};

const pascalToKebab$1 = (string) =>
  string.replace(/[\w]([A-Z])/g, function (m) {
    return m[0] + "-" + m[1].toLowerCase()
  });

const kebabToPascal$1 = (string) =>
  string.replace(/[\w]-([\w])/g, function (m) {
    return m[0] + m[2].toUpperCase()
  });

const parseStyles = (value) => {
  let type = typeof value;

  if (type === "string")
    return value.split(";").reduce((o, value) => {
      const [k, v] = value.split(":").map((v) => v.trim());
      if (k) o[k] = v;
      return o
    }, {})

  if (type === "object") return value

  return {}
};

const joinStyles = (value) =>
  Object.entries(value)
    .map(([k, v]) => `${k}: ${v};`)
    .join(" ");

const convertStyles = (o) =>
  Object.keys(o).reduce((a, k) => {
    a[pascalToKebab$1(k)] = o[k];
    return a
  }, {});

const applyAttribute$1 = (node, name, value) => {
  if (name === "style") {
    value = joinStyles(
      convertStyles({
        ...parseStyles(node.getAttribute("style")),
        ...parseStyles(value),
      })
    );
  } else if (name === "class") {
    switch (typeOf(value)) {
      case "Array":
        value = value.join(" ");
        break
      case "Object":
        value = Object.keys(value)
          .reduce((a, k) => {
            if (value[k]) a.push(k);
            return a
          }, [])
          .join(" ");
        break
    }
  } else if (!isPrimitive(value)) {
    return (node[kebabToPascal$1(name)] = value)
  }

  name = pascalToKebab$1(name);

  if (typeof value === "boolean") {
    if (name.startsWith("aria-")) {
      value = "" + value;
    } else if (value) {
      value = "";
    }
  }

  let current = node.getAttribute(name);

  if (value === current) return

  if (typeof value === "string" || typeof value === "number") {
    node.setAttribute(name, value);
  } else {
    node.removeAttribute(name);
  }
};

const handler = ({ path, identifier, key, index, i, k }) => ({
  get(target, property) {
    let x = getValueAtPath(path, target);

    // x === the collection

    if (property === identifier) {
      for (let n in x) {
        let v = x[n];
        if (key) {
          if (v[key] === k) return v
        } else {
          if (n == i) return v
        }
      }
    }

    if (property === index) {
      for (let n in x) {
        let v = x[n];
        if (key) {
          if (v[key] === k) return n
        } else {
          if (n == i) return n
        }
      }
    }

    let t = key ? x.find((v) => v[key] === k) : x?.[i];
    if (t?.hasOwnProperty?.(property)) return t[property]

    return Reflect.get(...arguments)
  },
  set(target, property, value) {
    let x = getValueAtPath(path, target);
    let t = key ? x.find((v) => v[key] === k) : x[i];
    if (t && !isPrimitive(t)) {
      t[property] = value;
      return true
    }

    return Reflect.set(...arguments)
  },
});

const createContext = (v = []) => {
  let context = v;
  return {
    get: () => context,
    push: (v) => context.push(v),
    wrap: (state) => {
      return context.reduce(
        (target, ctx) => new Proxy(target, handler(ctx)),
        state
      )
    },
  }
};

const convertToTemplate = (node) => {
  let ns = node.namespaceURI;

  if (ns.endsWith("/svg")) {
    let tpl = document.createElementNS(ns, "defs");
    tpl.innerHTML = node.outerHTML;
    node.parentNode.replaceChild(tpl, node);
    return tpl
  } else {
    if (node.nodeName === "TEMPLATE") return node

    let tpl = document.createElement("template");

    tpl.innerHTML = node.outerHTML;
    node.parentNode.replaceChild(tpl, node);

    return tpl
  }
};

function nextWord(css, count) {
  return css.slice(count - 1).split(/[\s+|\n+|,]/)[0]
}

function nextOpenBrace(css, count) {
  let index = css.slice(count - 1).indexOf("{");
  if (index > -1) {
    return count + index
  }
}

function prefixSelectors(prefix, css) {
  let insideBlock = false;
  let look = true;
  let output = "";
  let count = 0;
  let skip = false;

  for (let char of css) {
    if (char === "@" && nextWord(css, count + 1) === "@media") {
      skip = nextOpenBrace(css, count);
    }

    if (skip) {
      if (skip === count) skip = false;
    }

    if (!skip) {
      if (char === "}") {
        insideBlock = false;
        look = true;
      } else if (char === ",") {
        look = true;
      } else if (char === "{") {
        insideBlock = true;
      } else if (look && !insideBlock && !char.match(/\s/)) {
        let w = nextWord(css, count + 1);

        // console.log({ w })

        if (
          w !== prefix &&
          w.charAt(0) !== "@" &&
          w.charAt(0) !== ":" &&
          w.charAt(0) !== "*" &&
          w !== "html" &&
          w !== "body"
        ) {
          output += prefix + " ";
        }
        look = false;
      }
    }
    output += char;
    count += 1;
  }

  return output
}

function appendStyles(name, css) {
  if (document.querySelector(`style#${name}`)) return

  const el = document.createElement("style");
  el.id = name;
  el.innerHTML = prefixSelectors(name, css);
  document.head.appendChild(el);
}

const partials = {};

const partial = (name, html, css) => {
  if (css) appendStyles(name, css);
  partials[name.toUpperCase()] = html;
};

const render = (
  target,
  { getState, dispatch, refs },
  template,
  updatedCallback,
  beforeMountCallback
) => {
  let observer = () => {
    let subscribers = new Set();
    return {
      publish: (cb) => {
        for (let fn of subscribers) {
          fn();
        }
        cb?.();
      },
      subscribe(fn) {
        subscribers.add(fn);
      },
    }
  };

  const createSubscription = {
    [TEXT]: ({ value, node, context }, { getState }) => {
      return {
        handler: () => {
          let state = context ? context.wrap(getState()) : getState();
          let a = node.textContent;
          let b = getValueFromParts(state, getParts(value));

          if (a !== b) node.textContent = b;
        },
      }
    },
    [ATTRIBUTE]: ({ value, node, name, context }, { getState }) => {
      return {
        handler: () => {
          let state = context ? context.wrap(getState()) : getState();
          let b = getValueFromParts(state, getParts(value));

          applyAttribute$1(node, name, b);

          if (node.nodeName === "OPTION") {
            let path = node.parentNode.getAttribute("name");
            let selected = getValueAtPath(path, state);
            node.selected = selected === b;
          }
        },
      }
    },
    [INPUT]: ({ node, path, context }, { getState, dispatch }) => {
      let eventType = ["radio", "checkbox"].includes(node.type)
        ? "click"
        : "input";

      node.addEventListener(eventType, () => {
        let value =
          node.getAttribute("type") === "checkbox" ? node.checked : node.value;

        if (value.trim?.().length && !isNaN(value)) value = +value;

        if (context) {
          let state = context.wrap(getState());
          setValueAtPath(path, value, state);
          dispatch({
            type: "MERGE",
            payload: state,
          });
        } else {
          dispatch({
            type: "SET",
            payload: {
              name: path,
              value,
              context,
            },
          });
        }
      });

      return {
        handler: () => {
          let state = context ? context.wrap(getState()) : getState();
          updateFormControl(node, getValueAtPath(path, state));
        },
      }
    },
    [EVENT]: (
      { node, eventType, actionType, context },
      { dispatch, getState }
    ) => {
      node.addEventListener(eventType, (event) => {
        let action = {
          type: actionType,
          event,
          scope: context ? context.wrap(getState()) : getState(),
        };

        dispatch(action);
      });
      return {
        handler: () => {},
      }
    },
    [CONDITIONAL]: ({ node, expression, context, map }, { getState }) => {
      return {
        handler: () => {
          let state = context ? context.wrap(getState()) : getState();
          let shouldMount = getValueFromParts(
            state,
            getParts(wrapToken(expression))
          );
          let isMounted = node.dataset.mounted === "1";

          if (shouldMount && !isMounted) {
            // MOUNT
            let frag = fragmentFromTemplate(node);
            walk(frag.firstChild, bindAll(map, null, context));
            node.after(frag);
            node.dataset.mounted = "1";
          } else if (!shouldMount && isMounted) {
            // UNMOUNT
            node.nextSibling.remove();
            delete node.dataset.mounted;
          }
        },
        // pickupNode,
      }
    },
    [REPEAT]: (
      { node, context, map, path, identifier, index, key, blockIndex, hydrate },
      { getState }
    ) => {
      let oldValue;
      node.$t = blockIndex - 1;

      let pickupNode = node.nextSibling;

      const initialiseBlock = (rootNode, i, k, exitNode) => {
        walk(
          rootNode,
          multi(
            (node) => {
              if (node === exitNode) return false
            },
            bindAll(
              map,
              hydrate,
              createContext(
                (context?.get() || []).concat({
                  path,
                  identifier,
                  key,
                  index,
                  i,
                  k,
                })
              )
            ),
            (child) => (child.$t = blockIndex)
          )
        );
      };

      function firstChild(v) {
        return (v.nodeType === v.DOCUMENT_FRAGMENT_NODE && v.firstChild) || v
      }

      const createListItem = (datum, i) => {
        let k = datum[key];
        let frag = fragmentFromTemplate(node);

        initialiseBlock(firstChild(frag), i, k);
        return frag
      };

      if (hydrate) {
        let x = getValueAtPath(path, getState());
        let blocks = getBlocks(node);

        blocks.forEach((block, i) => {
          let datum = x[i];
          let k = datum?.[key];
          initialiseBlock(block[0], i, k, last(block).nextSibling);
        });

        pickupNode = last(last(blocks))?.nextSibling;
      }

      return {
        handler: () => {
          let state = context ? context.wrap(getState()) : getState();

          const newValue = Object.entries(getValueAtPath(path, state) || []);
          const delta = compareKeyedLists(key, oldValue, newValue);

          if (delta) {
            updateList(node, delta, newValue, createListItem);
          }
          oldValue = newValue.slice(0);
        },
        pickupNode,
      }
    },
  };

  const mediator = () => {
    const o = observer();
    return {
      bind(v) {
        let s = createSubscription[v.type](v, { getState, dispatch });

        o.subscribe(s.handler);
        return s
      },

      update(cb) {
        return o.publish(cb)
      },
    }
  };

  const { bind, update } = mediator();

  let blockCount = 0;

  function compareAttributes(a, b) {
    if (b.name === ":if") {
      return -1
    }
    if (a.name === ":if") {
      return 1
    }
    return 0
  }

  const parse = (frag) => {
    let index = 0;
    let map = {};

    walk(frag, (node) => {
      let x = [];
      let pickupNode;

      switch (node.nodeType) {
        case node.TEXT_NODE: {
          let value = node.textContent;
          if (hasMustache(value)) {
            x.push({
              type: TEXT,
              value: value.trim(),
            });
          }
          break
        }
        case node.ELEMENT_NODE: {
          if (node.nodeName in partials) {
            node.innerHTML = partials[node.nodeName];
            break
          }

          let each = parseEach(node);

          if (each) {
            let ns = node.namespaceURI;
            let m;

            node.removeAttribute(":each");
            node = convertToTemplate(node);

            if (ns.endsWith("/svg")) {
              m = parse(node.firstChild);
            } else {
              m = parse(node.content.firstChild);
            }

            pickupNode = node.nextSibling;

            x.push({
              type: REPEAT,
              map: m,
              blockIndex: blockCount++,
              ...each,
            });

            break
          }

          let attrs = [...node.attributes].sort(compareAttributes);
          let i = attrs.length;
          while (i--) {
            let { name, value } = attrs[i];

            if (
              name === ":name" &&
              value &&
              (node.nodeName === "INPUT" ||
                node.nodeName === "SELECT" ||
                node.nodeName === "TEXTAREA")
            ) {
              x.push({
                type: INPUT,
                path: value,
              });

              node.removeAttribute(name);
              node.setAttribute("name", value);
            } else if (name === ":if") {
              node.removeAttribute(name);
              node = convertToTemplate(node);
              pickupNode = node.nextSibling;

              x.push({
                type: CONDITIONAL,
                expression: value,
                map: parse(node.content?.firstChild || node.firstChild),
              });
            } else if (name === ":ref") {
              node.removeAttribute(name);
              node.setAttribute("ref", value);
              Object.defineProperty(refs, value, {
                get() {
                  return document.querySelector(`[ref="${value}"]`)
                },
              });
            } else if (name.startsWith(":on")) {
              node.removeAttribute(name);
              let eventType = name.split(":on")[1];
              x.push({
                type: EVENT,
                eventType,
                actionType: value,
              });
            } else if (name.startsWith(":")) {
              let prop = name.slice(1);

              let v = value || prop;

              if (!v.includes("{{")) v = `{{${v}}}`;

              x.push({
                type: ATTRIBUTE,
                name: prop,
                value: v,
              });
              node.removeAttribute(name);
            }
          }
        }
      }
      if (x.length) {
        map[index] = x;
      }
      index++;
      return pickupNode
    });

    return map
  };

  const multi =
    (...fns) =>
    (...args) => {
      for (let fn of fns) {
        let v = fn(...args);
        if (v === false) return false
      }
    };

  const bindAll = (bMap, hydrate = 0, context) => {
    let index = 0;
    return (node) => {
      let k = index;
      let p;

      if (k in bMap) {
        bMap[k].forEach((v) => {
          let x = bind({
            ...v,
            node,
            context,
            hydrate,
          });
          p = x.pickupNode;
        });
        node.$i = index;
      }

      index++;
      return p
    }
  };

  let frag = fragmentFromTemplate(template);
  let map = parse(frag);
  let ds = getDataScript(target);

  if (ds) {
    walk(target, bindAll(map, 1));
  } else {
    walk(frag, bindAll(map));
    beforeMountCallback?.(frag);
    update();
    target.prepend(frag);
  }

  return debounce(() => update(updatedCallback))
};

const childNodes = (node) => {
  let frag = document.createDocumentFragment();
  while (node.firstChild) {
    frag.appendChild(node.firstChild);
  }
  return frag
};

const mergeSlots = (targetNode, sourceNode) => {
  let namedSlots = sourceNode.querySelectorAll("slot[name]");

  namedSlots.forEach((slot) => {
    let name = slot.attributes.name.value;
    let node = targetNode.querySelector(`[slot="${name}"]`);
    if (!node) {
      slot.parentNode.replaceChild(childNodes(slot), slot);
      return
    }
    node.removeAttribute("slot");
    slot.parentNode.replaceChild(node, slot);
  });

  let defaultSlot = sourceNode.querySelector("slot:not([name])");

  if (defaultSlot) {
    defaultSlot.parentNode.replaceChild(
      childNodes(targetNode.innerHTML.trim() ? targetNode : defaultSlot),
      defaultSlot
    );
  }
};

function serialise(node, state) {
  let ds = getDataScript(node) || createDataScript(node);
  ds.innerText = JSON.stringify(state);
}

const define = (name, factory, template, css) => {
  if (customElements.get(name)) return

  if (css) appendStyles(name, css);

  customElements.define(
    name,
    class extends HTMLElement {
      async connectedCallback() {
        if (!this.initialised) {
          let config = factory(this);

          if (config instanceof Promise) config = await config;

          let { subscribe, shadow, observe = [] } = config;

          this.connectedCallback = config.connectedCallback;
          this.disconnectedCallback = config.disconnectedCallback;

          const ds = getDataScript(this);

          const { dispatch, getState, onChange, updated, refs } = configure(
            {
              ...config,
              state: ds ? JSON.parse(ds.innerText) : config.state,
            });

          this.store = { dispatch, getState };

          let state = getState();

          let observedAttributes = observe.map(pascalToKebab);

          let sa = this.setAttribute;
          this.setAttribute = (k, v) => {
            if (observedAttributes.includes(k)) {
              let { name, value } = attributeToProp(k, v);

              dispatch({
                type: "SET",
                payload: { name, value },
              });
            }
            sa.apply(this, [k, v]);
          };
          let ra = this.removeAttribute;
          this.removeAttribute = (k) => {
            if (observedAttributes.includes(k)) {
              let { name, value } = attributeToProp(k, null);
              dispatch({
                type: "SET",
                payload: { name, value },
              });
            }
            ra.apply(this, [k]);
          };

          observedAttributes.forEach((name) => {
            let property = attributeToProp(name).name;
            let value;

            if (this.hasAttribute(name)) {
              value = this.getAttribute(name);
            } else {
              value = this[property] || state[property];
            }

            Object.defineProperty(this, property, {
              get() {
                return getState()[property]
              },
              set(v) {
                dispatch({
                  type: "SET",
                  payload: { name: property, value: v },
                });
                if (isPrimitive(v)) {
                  applyAttribute(this, property, v);
                }
              },
            });

            this[property] = value;
          });

          let beforeMountCallback;

          if (shadow) {
            this.attachShadow({
              mode: shadow,
            });
          } else {
            beforeMountCallback = (frag) => mergeSlots(this, frag);
          }

          onChange(
            render(
              this.shadowRoot || this,
              { getState, dispatch, refs },
              template,
              () => {
                const state = getState();

                serialise(this, state);
                observe.forEach((k) => {
                  let v = state[k];
                  if (isPrimitive(v)) applyAttribute(this, k, v);
                });
                subscribe?.(getState());
                updated();
              },
              beforeMountCallback
            )
          );

          if (!ds) serialise(this, getState());
        }

        if (!this._C) this.connectedCallback?.(this.store);
        this._C = false;
      }
      disconnectedCallback() {
        this._C = true;
        requestAnimationFrame(() => {
          if (this._C) this.disconnectedCallback?.();
          this._C = false;
        });
      }
    }
  );
};

export { define, partial };
